/* ********************************************************************
 * Copyright (c) 2019 Biological Modeling and Visualization Laboratory,
 * University of Calgary. All rights reserved.
 * ********************************************************************/

#include <cmath>
#include <lpfgall.h>
#include "bspline.h"
#include "markers.h"


#define CURVE_SCALE 	3.0		// scaling factor for the graphically-defined apex profile
#define SPHERE_RADIUS 	0.007	
#define LINE_WIDTH      0.007 

#define dns 			0.001 	// normalized arc-length distance betwen consecutive points
#define DALPHA			0.01 	// interpolation step


const int NUM_CNTRS = 11; // number of contours

/* Interpolate between the contour curves */
// Specifically, given:
// - graphically defined profile (contour) curves,
//   both parametrized by normalized arc-lenth coordinate ns;
// - a weighting of the curves 
// - a value of coordinate ns in [0,1]
// return position of the interpolated point.

V3d hatProfileXYZ(double ns)
{
    extern BSplineBasisFunc contourWeights;
    extern BSplineBasisFunc primordiaMarkerWeights;
    extern vector<vector<double>> primordiaMarkers;
    
    primordiaMarkerWeights.computeFuncs(ns);

    V3d v(0,0,0);
    for (int i = 1; i <= NUM_CNTRS; i++) {
        double newNS = 0.;
        for (int j = 0; j < primordiaMarkers.size(); j++) {
            newNS += primordiaMarkerWeights.valueAtFunc(j) * primordiaMarkers[j][i-1];
        }

		v += contourWeights.valueAtFunc(i-1) * ((V3d) curveXYZ(i,newNS));
	}
    return CURVE_SCALE * v;
}

V3d pure_profileXYZ(double ns) // C(t,s) in description
{
    extern BSplineBasisFunc contourWeights;
	V3d res(0,0,0);
	for (int i = 1; i <= NUM_CNTRS; i++) {
		res += contourWeights.valueAtFunc(i-1) * ((V3d) curveXYZ(i,ns)); // curveXYZ(k,s) is C_k(s) in description
	}
	
	return CURVE_SCALE * res;
}


double newNS(int markerIndex)
{
    extern BSplineBasisFunc markerWeights;
	extern vector<vector<double>> primordiaMarkers;
	double new_ns = 0.;	
	for (int i = 0; i < NUM_CNTRS; i++) {
		new_ns += markerWeights.valueAtFunc(i) * primordiaMarkers[markerIndex][i];
	}	
	return new_ns;
}

V3d profileXYZ(int markerIndex)
{
	double new_ns = newNS(markerIndex);
	return pure_profileXYZ(new_ns);
}


// interpolation functions for active ring
double active_ring_newNS() // s^{gen}(t) in description
{
    extern BSplineBasisFunc markerWeights;
    extern vector<double> activeRingMarkers;
    double new_ns = 0.;
    for (int i = 0; i < activeRingMarkers.size(); i++) {
        new_ns += markerWeights.valueAtFunc(i) * activeRingMarkers[i];
    }
	return new_ns;
}

V3d active_ring_profileXYZ() // G(t) in description
{
	double new_ns = active_ring_newNS();	
	return pure_profileXYZ(new_ns);
}


module Recompute();	
module DrawAll(double);
module DrawData();
module DrawBackground();

derivation length: 2;

Axiom:	Recompute();

const int CONTOUR_WEIGHTS_SPLINE_DEGREE = 2;
BSplineBasisFunc contourWeights(NUM_CNTRS, CONTOUR_WEIGHTS_SPLINE_DEGREE, 1);

const int MARKER_WEIGHTS_SPLINE_DEGREE = 2;
BSplineBasisFunc markerWeights(NUM_CNTRS, MARKER_WEIGHTS_SPLINE_DEGREE, 1);

void resetWeights (double alpha) {
	contourWeights.computeFuncs(alpha);
	markerWeights.computeFuncs(alpha);
}

BSplineBasisFunc primordiaMarkerWeights(0,MARKER_WEIGHTS_SPLINE_DEGREE,1);

vector<double> activeRingMarkers;
vector<vector<double>> primordiaMarkers;

Start: {
	// scale the first contour by 19% (estimated by overlapping images)
	curveScale(1, 0.19, 0.19, 0.19);
}

production:

Recompute() :
{
	Printf ("Recompute...");

    activeRingMarkers.clear();

    int maxLandmarks = myCurveMarkers(NUM_CNTRS)-1; // -1 for active ring marker

    primordiaMarkers = vector<vector<double>> (maxLandmarks+1,vector<double>(NUM_CNTRS,0.0)); // + 1 for s = 1

    // make an array of active ring markers and another of the primordia markers
   	for (int cntr = 1; cntr <= NUM_CNTRS; cntr++) {
        int numMarkers = myCurveMarkers(cntr);
        activeRingMarkers.push_back(myCurveMarker(cntr,numMarkers));

        primordiaMarkers[0][cntr-1] = 1.0;
        for (int i = 1; i < numMarkers; i++)
            primordiaMarkers[i][cntr-1] = myCurveMarker(cntr,i);
    } 

    // propagate landmarks to earlier contours
    for (int i = 0; i < primordiaMarkers.size(); i++) {
        // find the first non-zero entry
        int zeroIndex = primordiaMarkers[i].size() - 1;
        while(primordiaMarkers[i][zeroIndex-1] > 0. && zeroIndex-1 >= 0)
               --zeroIndex;

        // if there are zero entries
        if (zeroIndex-1 > 0) {
            // find the ratio of expansion between the new landmark and existing one
            double ratio = primordiaMarkers[i][zeroIndex] / primordiaMarkers[i-1][zeroIndex];
            // propagate the ratio to the earlier contours
            for (int k = 0; k < zeroIndex; k++) 
                    primordiaMarkers[i][k] = ratio * primordiaMarkers[i-1][k];
        }
    }

    // add landmarks at s = 0 for each contour
    primordiaMarkers.push_back(vector<double>(NUM_CNTRS,0.0));

    // create the B-spline weights for the primordia markers
    primordiaMarkerWeights = BSplineBasisFunc(primordiaMarkers.size(), MARKER_WEIGHTS_SPLINE_DEGREE, 1);

	Printf ("Done.\n");

	produce DrawBackground() DrawData() DrawAll(DALPHA);
}

DrawAll(dalpha) :
{
	nproduce SetWidth(LINE_WIDTH*2.) SetColor(1);

	// draw interpolated contours
	double alpha = 0.;
	while (alpha < 1.) {
		resetWeights(alpha);
        nproduce MoveTo3d(hatProfileXYZ(0.));
		for (double ns= dns; ns<1.; ns+= dns) {
			nproduce LineTo3d(hatProfileXYZ(ns));
        }
        nproduce LineTo3d(hatProfileXYZ(1.));
		alpha += 30.0 * dalpha * func(output_dalpha,alpha);
	}

	// draw lines connecting markers across contours
	nproduce SetWidth(SPHERE_RADIUS*0.8 * 2.);
	int colour = 2;
	for (int i = 0; i < primordiaMarkers.size(); ++i) {

        int numMarkers = primordiaMarkers[i].size();

        // draw starting point
        resetWeights(0.);
        V3d v = profileXYZ(i) + V3d(0.,0.,0.01);
        nproduce SetColor(colour) MoveTo3d(v) ;

		// line to next marker (increment alpha by dalpha)
        // change colour when the line intersects the active ring line
		double nsAR = myCurveMarker(1,myCurveMarkers(1));
		double alpha = 0.;
		while (alpha < 1.) {
			resetWeights(alpha);
			// change colour once marker passes the active ring
			double new_nsAR = active_ring_newNS();	
			if (newNS(i) >= new_nsAR)
				nproduce SetColor(colour);
			else
				nproduce SetColor(colour+5);

			v = profileXYZ(i) + V3d(0.,0.,0.01);
			nproduce LineTo3d(v) ;
			alpha += 1.0 * dalpha * func(output_dalpha,alpha);
		}

        // draw ending point
        resetWeights(1.);
        v = profileXYZ(i) + V3d(0.,0.,0.01);
        nproduce LineTo3d(v) Sphere(SPHERE_RADIUS*0.8);
	}

	// draw the interpolated active ring marker
	nproduce SetWidth(LINE_WIDTH*1.3 * 2.);

    resetWeights(0.);
    V3d v = active_ring_profileXYZ() + V3d(0.,0.,0.015);
	nproduce SetColor(3) MoveTo3d(v) Sphere(SPHERE_RADIUS*1.3);

	for (double alpha = 0.; alpha < 1.; alpha += dalpha) {
		resetWeights(alpha);
        v = active_ring_profileXYZ() + V3d(0.,0.,0.015);
		nproduce LineTo3d(v) Sphere(SPHERE_RADIUS*1.3);
	}

	resetWeights(1.);
    v = active_ring_profileXYZ() + V3d(0.,0.,0.015);
	nproduce LineTo3d(v) Sphere(SPHERE_RADIUS*1.3);

 	produce;
}

DrawData() :
{
	nproduce SetWidth(LINE_WIDTH*1.1*2.) ;

	for (int cntr = 1; cntr <= NUM_CNTRS; ++cntr) {
	
		nproduce SB SetColor(4) MoveTo3d(CURVE_SCALE * (V3d) curveXYZ(cntr,0.)) Right(90) StartGC;
		for (double ns=dns; ns<1.; ns+= dns)
			nproduce LineTo3d(CURVE_SCALE * (V3d) curveXYZ(cntr,ns) );
		nproduce LineTo3d(CURVE_SCALE * (V3d) curveXYZ(cntr,1.)) EndGC EB;

		// draw marker for the bracts on curve
		nproduce SetColor(2);
		int numMarkers = myCurveMarkers(cntr);
		for (int i = 0; i < numMarkers-1; i++) {
			double ns = myCurveMarker(cntr,i+1);
			V3d v = CURVE_SCALE * (V3d) curveXYZ(cntr,ns) + V3d(0.,0.,0.01);
			nproduce MoveTo3d(v) Sphere(SPHERE_RADIUS*1.4 * 1.5);
		}

		// draw the active ring marker on curve
		double ns = myCurveMarker(cntr,numMarkers);
		V3d v = CURVE_SCALE * (V3d) curveXYZ(cntr,ns);
		nproduce SetColor(6) MoveTo3d(v) ;
	}

	produce;
}

DrawBackground():
{
	const double SCALE = 0.0034;
	produce SB SetColor(12) MoveTo(-1293.*SCALE*0.5,0.,-0.01) CurrentTexture(0) Surface(0,SCALE) EB;
}

