// Just an attempt to encapsulate all the CGAL stuff to reduce compilation time

#include <CGAL/Exact_predicates_inexact_constructions_kernel.h> 
#include <CGAL/Advancing_front_surface_reconstruction.h>
#include <CGAL/Surface_mesh.h>
#include <CGAL/array.h>
#include <vector>
#include <cassert>

#include <lintrfc.h>

#include "voronoi.h"


typedef CGAL::Exact_predicates_inexact_constructions_kernel Kernel;
typedef Kernel::Point_3  Point_3;
typedef CGAL::Surface_mesh<Point_3> CGAL_TriangleMesh;
typedef CGAL_TriangleMesh::Vertex_index vertex_descriptor;

struct Construct{
  CGAL_TriangleMesh& mesh;
  template < typename PointIterator>
  Construct(CGAL_TriangleMesh& mesh,PointIterator b, PointIterator e)
    : mesh(mesh)
  {
    for(; b!=e; ++b){
      vertex_descriptor v = add_vertex(mesh);
      mesh.point(v) = *b;
    }
  }
  Construct& operator=(const CGAL::cpp11::array<std::size_t,3> f)
  {
    typedef CGAL_TriangleMesh::vertices_size_type size_type;
    mesh.add_face(vertex_descriptor(static_cast<size_type>(f[0])),
                  vertex_descriptor(static_cast<size_type>(f[1])),
                  vertex_descriptor(static_cast<size_type>(f[2])));
    return *this;
  }
  Construct&
  operator*() { return *this; }
  Construct&
  operator++() { return *this; }
  Construct
  operator++(int) { return *this; }
};


// Global variables ...
CGAL_TriangleMesh mesh; // the Delaunay triangle mesh

// extern variables in lsystem.l (would be better to return const reference)
std::vector<V3d> mesh_vertices;
std::vector<std::vector<unsigned int>> mesh_vertex_neighbours;
std::vector<std::vector<VoronoiCoordinates>> voronoi_cells; // voronoi cells' vertices



void CGAL_ComputeVoronoiCells (const std::vector<V3d> &inPts, const V3d &max_pt,
		const V3d &min_pt, double apical_y)
{
	// copy primordia (x,y,z) positions to CGAL type 
	std::vector<Point_3> points;
	for (int i = 0; i< inPts.size(); i++) {
		points.push_back(Point_3(inPts[i].x, inPts[i].y, inPts[i].z));
	}

	// generate ring of points near the bottom of the receptacle
	for (double phi = 0; phi < 2.*M_PI; phi += 2.*M_PI/20.) {
		V3d primordium_pos (max_pt.x*cos(phi), max_pt.y, -max_pt.x*sin(phi)); 
		points.push_back(Point_3(primordium_pos.x,primordium_pos.y,primordium_pos.z));
	}
	points.push_back(Point_3(0.0,0.0,0.0));
	
	// generate ring of points near the top of the receptacle
	if (apical_y > 0.0) {
		for (double phi = 0; phi < 2.*M_PI; phi += 2.*M_PI/20.) {
			V3d primordium_pos (min_pt.x*cos(phi), min_pt.y, -min_pt.x*sin(phi)); 
			points.push_back(Point_3(primordium_pos.x,primordium_pos.y,primordium_pos.z));
		}
		points.push_back(Point_3(0.,apical_y,0.));
	}

	// call the CGAL function to construct a Delaunay triangle mesh
	mesh.clear();
	Construct construct(mesh,points.begin(),points.end());
  	CGAL::advancing_front_surface_reconstruction(points.begin(),points.end(),construct);

	// compute the Voronoi diagram given the triangle mesh
	voronoi_cells.clear();
	mesh_vertices.clear();
	mesh_vertex_neighbours.clear();
	for(vertex_descriptor vd : mesh.vertices()){

		// if this vertex is from the ring of points, skip it
		if (vd >= inPts.size())
			continue;

		// iterators for neighbours of the current vertex, vd
		CGAL_TriangleMesh::Halfedge_index heIndex = mesh.halfedge(vd);
		// not sure why, but sometimes heIndex is incorrect:
		if (heIndex > mesh.number_of_halfedges()) {
			fprintf (stderr, "Skipping invalid halfedge!\n");
			continue;
		}
		CGAL::Vertex_around_target_circulator<CGAL_TriangleMesh> vbegin(heIndex,mesh);
		CGAL::Vertex_around_target_circulator<CGAL_TriangleMesh> vend(vbegin);

		// get the 3D point
		const Point_3 & p0 = mesh.point(vd);
		V3d v0 (CGAL::to_double(p0.x()),CGAL::to_double(p0.y()),CGAL::to_double(p0.z()));
		mesh_vertices.push_back(v0);

		std::vector<unsigned int> vertex_neighbours;

		if (CGAL::circulator_size(vbegin) > 2) { // need at least 2 neighbouring vertices
			std::vector<VoronoiCoordinates> voronoi_cell;
			do {
				// get two vertices around this vertex (v0)
				vertex_descriptor vd1 = *vbegin;
				vertex_descriptor vd2 = *(++vbegin); // advance vbegin here!

				if (vd1 < inPts.size())
					vertex_neighbours.push_back(static_cast<unsigned int>(vd1));

				// get the two corresponding 3D points
				const Point_3 & p1 = mesh.point(vd1);
				const Point_3 & p2 = mesh.point(vd2);

				// compute the circumcenter of the triangle
				Point_3 c = CGAL::circumcenter(p0,p1,p2);

				// add to voronoi cell
				VoronoiCoordinates v;
				v.pos = V3d(CGAL::to_double(c.x()),CGAL::to_double(c.y()),CGAL::to_double(c.z()));
				v.syncCylindrical();
				// arc length coordinate computed in lsystem.l
				//v.ns = GetContourSampleInv(v.r,v.h);
				//v.syncCoordinates(); // given ns and phi, compute r,h and pos
				voronoi_cell.push_back(v);

			} while(vbegin != vend);

			voronoi_cells.push_back(voronoi_cell);
		}
		mesh_vertex_neighbours.push_back(vertex_neighbours);
    	}

	return;
}

